<?php

/**
 *  @file
 *  This includes internal form calls for various admin screens. They are used rarely, and only called from within
 *  admin setup screens, so are only loaded when needed.
 */

/**
 *  this returns a cached menu, called from rpg_menu()
 */
function _rpg_menu_cache() {
  $items = array();
  $admin_access = user_access('administer rpg');

  // the main game playing page
  $items[] = array(
    'path' => 'rpg',
    'title' => t('RPG'),
    'callback' => 'rpg_page',
    'access' => ($admin_access || user_access('play rpg')),
  );
  $items[] = array(
    'path' => 'rpg/play',
    'title' => t('Play game'),
    'callback' => 'rpg_play',
    'access' => ($admin_access || user_access('play rpg')),
  );
  $items[] = array(
    'path' => 'rpg/select',
    'title' => t('Select character'),
    'callback' => 'rpg_character_select_page',
    'access' => ($admin_access || user_access('play rpg')),
  );
  // the page to create rpg objects
  $items[] = array(
    'path' => 'rpg/create',
    'title' => t('Create RPG Object'),
    'callback' => 'rpg_create_page',
    'access' => ($admin_access || user_access('create rpg object') || user_access('create pc rpg object')),
  );
  $items[] = array(
    'path' => 'rpg/edit',
    'title' => t('Edit RPG Object'),
    'callback' => 'rpg_edit_page',
    'access' => ($admin_access || user_access('edit rpg object') || user_access('edit own rpg object') || user_access('edit own pc rpg object')),
    // access controlled at rpg_edit_page
  );
  $items[] = array(
    'path' => 'rpg/view',
    'title' => t('View RPG Object'),
    'callback' => 'rpg_view_page',
    'access' => ($admin_access || user_access('play rpg')),
  );
  $items[] = array(
    'path' => 'admin/rpg',
    'title' => t('RPG configuration'),
    'description' => t('Adjust RPG configuration options.'),
    'position' => 'right',
    'weight' => -5,
    'callback' => 'system_admin_menu_block_page',
    'access' => $admin_access,
  );
  $items[] = array(
    'path' => 'admin/rpg/game',
    'title' => t('RPG Game'),
    'description' => t('Configure RPG game settings.'),
    'callback' => 'drupal_get_form',
    'callback arguments' => 'rpg_admin_game',
    'access' => $admin_access,
  );
  $items[] = array(
    'path' => 'admin/rpg/game/settings',
    'title' => t('Settings'),
    'description' => t('Configure RPG game settings.'),
    'weight' => -3,
    'type' => MENU_DEFAULT_LOCAL_TASK,
  );
  $items[] = array(
    'path' => 'admin/rpg/types',
    'title' => t('RPG Types'),
    'description' => t('Define and override RPG object types.'),
    'callback' => 'rpg_admin_types',
    'access' => $admin_access,
  );
  $items[] = array(
    'path' => 'admin/rpg/types/add',
    'title' => t('Add RPG Type'),
    'description' => t('Create a new RPG type.'),
    'callback' => 'drupal_get_form',
    'callback arguments' => 'rpg_admin_type_add',
    'access' => $admin_access,
    'type' => MENU_CALLBACK,
  );
  foreach(rpg_types() as $type) {
    $items[] = array(
      'path' => 'admin/rpg/types/type/' . $type['type'],
      'title' => $type['name'],
      'description' => t('Configure the @type RPG type.', array('@type' => $type['name'])),
      'callback' => 'drupal_get_form',
      'callback arguments' => array('rpg_admin_type_edit', $type['type']),
      'access' => $admin_access,
      'type' => MENU_CALLBACK,
    );
    $items[] = array(
      'path' => 'admin/rpg/types/type/' . $type['type'] . '/edit',
      'title' => t('Edit @type', array('@type' => $type['name'])),
      'description' => t('Configure the @type RPG type.', array('@type' => $type['name'])),
      'callback' => 'drupal_get_form',
      'callback arguments' => array('rpg_admin_type_edit', $type['type']),
      'access' => $admin_access,
      'type' => MENU_DEFAULT_LOCAL_TASK,
      'weight' => -5,
    );
    $items[] = array(
      'path' => 'admin/rpg/types/type/' . $type['type'] . '/attributes',
      'title' => t('Attributes'),
      'description' => t('Configure the attributes for the @type RPG type.', array('@type' => $type['name'])),
      'callback' => 'drupal_get_form',
      'callback arguments' => array('rpg_admin_type_attributes_form', $type['type']),
      'access' => $admin_access,
      'type' => MENU_LOCAL_TASK,
    );
    $items[] = array(
      'path' => 'admin/rpg/types/type/' . $type['type'] . '/actions',
      'title' => t('Actions'),
      'description' => t('Configure the actions for the @type RPG type.', array('@type' => $type['name'])),
      'callback' => 'drupal_get_form',
      'callback arguments' => array('rpg_admin_type_actions', $type['type']),
      'access' => $admin_access,
      'type' => MENU_LOCAL_TASK,
    );
    $items[] = array(
      'path' => 'admin/rpg/types/type/' . $type['type'] . '/events',
      'title' => t('Events'),
      'description' => t('Configure the events for the @type RPG type.', array('@type' => $type['name'])),
      'callback' => 'drupal_get_form',
      'callback arguments' => array('rpg_admin_type_events', $type['type']),
      'access' => $admin_access,
      'type' => MENU_LOCAL_TASK,
    );
    $items[] = array(
      'path' => 'admin/rpg/types/type/' . $type['type'] . '/attributes/add',
      'title' => t('Add attribute'),
      'description' => t('Add an attribute for the @type RPG type.', array('@type' => $type['name'])),
      'callback' => 'drupal_get_form',
      'callback arguments' => array('rpg_admin_type_attribute_add', $type['type']),
      'access' => $admin_access,
      'type' => MENU_CALLBACK,
    );
    $items[] = array(
      'path' => 'admin/rpg/types/type/' . $type['type'] . '/attributes/edit',
      'title' => t('Edit attribute'),
      'description' => t('Edit an attribute for the @type RPG type.', array('@type' => $type['name'])),
      'callback' => 'drupal_get_form',
      'callback arguments' => array('rpg_admin_type_attribute_edit', $type['type']),
      'access' => $admin_access,
      'type' => MENU_CALLBACK,
    );
    $items[] = array(
      'path' => 'admin/rpg/types/type/' . $type['type'] . '/actions/add',
      'title' => t('Add action'),
      'description' => t('Add an action for the @type RPG type.', array('@type' => $type['name'])),
      'callback' => 'drupal_get_form',
      'callback arguments' => array('rpg_admin_type_action_add', $type['type']),
      'access' => $admin_access,
      'type' => MENU_CALLBACK,
    );
    $items[] = array(
      'path' => 'admin/rpg/types/type/' . $type['type'] . '/actions/edit',
      'title' => t('Edit action'),
      'description' => t('Edit an action for the @type RPG type.', array('@type' => $type['name'])),
      'callback' => 'drupal_get_form',
      'callback arguments' => array('rpg_admin_type_action_edit', $type['type']),
      'access' => $admin_access,
      'type' => MENU_CALLBACK,
    );
    $items[] = array(
      'path' => 'admin/rpg/types/type/' . $type['type'] . '/events/add',
      'title' => t('Add event'),
      'description' => t('Add an event for the @type RPG type.', array('@type' => $type['name'])),
      'callback' => 'drupal_get_form',
      'callback arguments' => array('rpg_admin_type_event_add', $type['type']),
      'access' => $admin_access,
      'type' => MENU_CALLBACK,
    );
    $items[] = array(
      'path' => 'admin/rpg/types/type/' . $type['type'] . '/events/edit',
      'title' => t('Edit event'),
      'description' => t('Edit an event for the @type RPG type.', array('@type' => $type['name'])),
      'callback' => 'drupal_get_form',
      'callback arguments' => array('rpg_admin_type_event_edit', $type['type']),
      'access' => $admin_access,
      'type' => MENU_CALLBACK,
    );
  }
  return $items;
}

function _rpg_admin_game() {
  $form = array();
  $form['user'] = array(
    '#type' => 'fieldset',
    '#title' => t('RPG User Settings'),
    '#collapsible' => true,
    '#collapsed' => true,
    '#description' => t('These settings affect how users are able to interact with the game. You will also need to modify the !access_rules for the site.', array('!access_rules' => l(t('RPG Module Access Rules'), 'admin/user/access'))),
  );
  $form['user']['rpg_user_max_pc'] = array(
    '#type' => 'textfield',
    '#title' => t('Maximum Number of Player Characters (PC\'s) Allowed per User'),
    '#default_value' => variable_get('rpg_user_max_pc', 0),
    '#size' => 7,
    '#description' => t('If this number is greater than 0, then users who have permission may only create up to that number of characters to play. This may be overridden per role below. Note that such users will also need to have enabled the \'play rpg\' access setting, and probably the appropriate create permissions for at least one rpg actor type of node (unless such characters are otherwise provided for them).'),
  );
  $form['user']['user_max_pc_roles'] = array(
    '#type' => 'fieldset',
    '#title' => t('Maximum Number of PC\'s Allowed per User by Role'),
    '#collapsible' => true,
    '#collapsed' => true,
    '#description' => t('If the number next to a role is greater than 0, then users with that role may only create up to that number of characters to play, overriding the above setting. A user with multiple roles will be assigned the largest number that overrides the global setting. Note that user roles with the access permission of \'Administer RPG Characters\' will always be able to create new characters, regardless of these settings.'),
  );
  foreach (user_roles() as $rid => $role) {
    $form['user']['user_max_pc_roles']['rpg_user_max_pc_role_' . $rid] = array(
      '#type' => 'textfield',
      '#title' => t('Maximum Number of PC\'s Allowed for @role Users', array('@role' => $role)),
      '#default_value' => variable_get('rpg_user_max_pc_role_' . $rid, 0),
      '#size' => 7,
    );
  }

  $duration = variable_get('rpg_action_round_duration', RPG_ACTION_ROUND_DURATION);
  $increment = variable_get('rpg_action_round_calendar_increment', RPG_ACTION_ROUND_CALENDAR_INCREMENT);
  $max_points = variable_get('rpg_max_action_points', RPG_MAX_ACTION_POINTS);
  $recharge = variable_get('rpg_action_recharge_rate', RPG_ACTION_RECHARGE_RATE);
  $figured = (int)((24*60*60) / $duration / $recharge);
  $suggested = (int)((24*60*60) / $duration / $max_points);

  $form['action'] = array(
    '#type' => 'fieldset',
    '#title' => t('RPG Action Round Settings'),
    '#collapsible' => true,
    '#collapsed' => true,
    '#description' => t('These settings affect Action Rounds in the game. Players may generally only make one action per round of play.'),
  );
  $form['action']['rpg_action_round_duration'] = array(
    '#type' => 'textfield',
    '#title' => t('Duration of Action Rounds'),
    '#default_value' => $duration,
    '#size' => 7,
    '#description' => t('Specify how many real-time seconds that an Action Round will last. A player must make an action within this window for it to be considered. You must enter a minimum of one second.'),
  );
  $form['action']['rpg_max_action_points'] = array(
    '#type' => 'textfield',
    '#title' => t('Maximum Character Action Points'),
    '#default_value' => $max_points,
    '#size' => 7,
    '#description' => t('This will determine the base maximum number of actions that a character may perform before they must recharge their points. This may be further modified by character types or by in-game modifiers. '),
  );
  $form['action']['rpg_action_recharge_rate'] = array(
    '#type' => 'textfield',
    '#title' => t('Action Point Recharge Rate'),
    '#default_value' => $recharge,
    '#size' => 7,
    '#description' => t('This determines how many rounds must pass for an action point to recharge. This may be further modified by character types or by in-game modifiers. At the current duration of @duration per round, with the current recharge rate of @rate, a character will recharge approximately @figured per real-time day. To recharge @points in approximately a day, specify the recharge rate as @suggested.', array('@duration' => format_plural($duration, '1 second', '@count seconds'), '@points' => format_plural($max_points, '1 point', '@count points'), '@rate' => format_plural($recharge, '1 round', '@count rounds'), '@figured' => format_plural($figured, '1 point', '@count points'), '@suggested' => $suggested)),
  );

  $form['calendar'] = array(
    '#type' => 'fieldset',
    '#title' => t('RPG Calendar Settings'),
    '#collapsible' => true,
    '#collapsed' => true,
    '#description' => t('These settings are for the in-game clock and calendar.'),
  );
  $form['calendar']['rpg_action_round_calendar_increment'] = array(
    '#type' => 'textfield',
    '#title' => t('Increment of Game Calendar by Action Round'),
    '#default_value' => $increment,
    '#size' => 7,
    '#description' => t('Specify how many game seconds pass per round of play in the game. Leave it at 0 for abstract or unspecified game time, or if the game calender will be incremented in some other method.'),
  );

  return $form;
}

/**
 *  this is the included callback function for /rpg/admin/types/type/[type]
 *  this form is the edit type form
 */
function _rpg_admin_type_form($type) {
  $form = array();
  $form['type'] = array(
    '#type' => 'fieldset',
    '#title' => t('Type Definition'),
    '#description' => $type['shortdesc'],
    '#collapsible' => true,
    '#collapsed' => false,
  );
  // if a module defines the type, then we lock the machine name
  $form['type']['type'] = array(
    '#type' => 'textfield',
    '#title' => t('Machine-Readable Name'),
    '#default_value' => $type['type'],
    '#description' => t('This is the machine-readable name for the type. It is used only internally. Machine names for types defined by modules may not be modified.'),
    '#disabled' => $type['lock'] ? true : false,
    '#required' => true,
  );
  $form['type']['name'] = array(
    '#type' => 'textfield',
    '#title' => t('Name'),
    '#default_value' => $type['name'],
    '#description' => t('This is the name of the type. It will be displayed during RPG object creation/editing, and on some object description screens.'),
  );
  $form['type']['parents'] = array(
    '#type' => 'fieldset',
    '#title' => t('Parents'),
    '#collapsible' => true,
    '#collapsed' => true,
  );
  $parents = array();
  foreach (rpg_types() as $parent) {
    $parents[$parent['type']] = $parent['name'];
  }
  $form['type']['parents']['parents'] = array(
    '#type' => 'select',
    '#multiple' => true,
    '#title' => t('Parents of @type', array('@type' => $type['name'])),
    '#options' => $parents,
    '#default_value' => $type['parents'],
    '#description' => t('These are the parents for this type. The type will inherit any attributes, actions, and events of the parent types, which may be overridden.'),
  );
  $form['type']['description'] = array(
    '#type' => 'fieldset',
    '#title' => t('Descriptions'),
    '#description' => t('These descriptions appear in various help screens, for instance when creating or editing objects.'),
    '#collapsible' => true,
    '#collapsed' => false,
  );
  $form['type']['description']['shortdesc'] = array(
    '#type' => 'textfield',
    '#title' => t('Short Description'),
    '#default_value' => $type['shortdesc'],
  );
  $form['type']['description']['longdesc'] = array(
    '#type' => 'textarea',
    '#title' => t('Long Description'),
    '#default_value' => $type['longdesc'],
  );

  $form['submit'] = array(
    '#type' => 'submit',
    '#value' => t('Submit'),
  );

  //type, name, module, parents, shortdesc, longdesc, data
  return $form;
}

function _rpg_admin_attribute_add_form($type, $attribute) {
  $form = array();

  $form['type'] = array(
    '#type' => 'value',
    '#value' => $type['type'],
  );

  $form['attribute'] = array(
    '#type' => 'textfield',
    '#title' => t('Attribute'),
    '#description' => t('This is the unique machine-readable name for the attribute, such as \'strength\' or \'hat_size\'. It may only contain alphanumeric characters and underscores.'),
    '#required' => true,
  );
  $form['name'] = array(
    '#type' => 'textfield',
    '#title' => t('Attribute name'),
    '#description' => t('This is the name of the attribute, such as \'Strength\' or \'Hat size\'.'),
    '#required' => true,
  );

  $classes = rpg_attribute_classes();
  // TODO: isn't there some php function to assign a subset of an array to another?
  $options = array();
  foreach ($classes as $class) {
    $options[$class['class']] = $class['label'];
  }
  $form['class'] = array(
    '#type' => 'radios',
    '#title' => t('Attribute class'),
    '#description' => t('Please select the class of attribute to define'),
    '#options' => $options,
    '#required' => true,
  );

  $form['submit'] = array(
    '#type' => 'submit',
    '#value' => t('Submit'),
  );
  return $form;
}

function _rpg_admin_action_add_form($type, $action) {
  $form = array();

  $form['type'] = array(
    '#type' => 'value',
    '#value' => $type['type'],
  );

  $form['action'] = array(
    '#type' => 'textfield',
    '#title' => t('Action'),
    '#description' => t('This is the unique machine-readable name for the action, such as \'take\' or \'throw_at\'. It may only contain alphanumeric characters and underscores.'),
    '#required' => true,
  );
  $form['name'] = array(
    '#type' => 'textfield',
    '#title' => t('Action name'),
    '#description' => t('This is the name of the action, such as \'Take\' or \'Throw at\'.'),
    '#required' => true,
  );

  $form['submit'] = array(
    '#type' => 'submit',
    '#value' => t('Submit'),
  );
  return $form;
}

function _rpg_admin_event_add_form($type, $event) {
  $form = array();

  $form['type'] = array(
    '#type' => 'value',
    '#value' => $type['type'],
  );

  $form['event'] = array(
    '#type' => 'textfield',
    '#title' => t('Event'),
    '#description' => t('This is the unique machine-readable name for the event, such as \'on_destroy\' or \'on_enter\'. It may only contain alphanumeric characters and underscores.'),
    '#required' => true,
  );
  $form['name'] = array(
    '#type' => 'textfield',
    '#title' => t('Event name'),
    '#description' => t('This is the name of the event, such as \'on exit\' or \'on create\'.'),
    '#required' => true,
  );

  $form['submit'] = array(
    '#type' => 'submit',
    '#value' => t('Submit'),
  );
  return $form;
}

function _rpg_admin_attribute_edit_form($type, $attribute) {
  $class = rpg_attribute_classes($attribute['class']);
  $form = array();

  $form['type'] = array(
    '#type' => 'value',
    '#value' => $type['type'],
  );
  $form['attribute'] = array(
    '#type' => 'value',
    '#value' => $attribute['attribute'],
  );
  $form['class'] = array(
    '#type' => 'value',
    '#value' => $attribute['class'],
  );
  $form['definition'] = array(
    '#type' => 'fieldset',
    '#title' => t('Attribute definition'),
    '#description' => t("The following items only affect the '%attribute' %class attribute of the %type type, and any types derived from it.", array('%attribute' => $attribute['name'], '%class' => $class['label'], '%type' => $type['name'])),
    '#collapsible' => true,
    '#collapsed' => false,
  );
  $form['definition']['name'] = array(
    '#type' => 'textfield',
    '#title' => t('Name'),
    '#default_value' => $attribute['name'],
  );
  $form['definition']['description'] = array(
    '#type' => 'textarea',
    '#title' => t('Description'),
    '#default_value' => $attribute['description'],
  );
  $form['definition']['get'] = array(
    '#type' => 'textfield',
    '#title' => t('Get'),
    '#description' => t("If it exists, this function will be called when the system attempts to get an object's value for this attribute. As the value returned will override the value stored in the database, this is generally used only for figured class attributes, but may be used on other occassions, or as a general way to override an inherited attribute with a figured value."),
    '#default_value' => $attribute['get'],
  );
  $form['definition']['verify'] = array(
    '#type' => 'textfield',
    '#title' => t('Verify'),
    '#description' => t("If it exists, this function will be called before setting an object's value for this attribute. If the function returns false, the value will not be set."),
    '#default_value' => $attribute['verify'],
  );
  $form['definition']['set'] = array(
    '#type' => 'textfield',
    '#title' => t('Set'),
    '#description' => t("If it exists, this function will be called after setting an object's value for this attribute. It is generally used to set other attributes dependent on this one, or to fire new events."),
    '#default_value' => $attribute['set'],
  );
  $form['definition']['default_value'] = array(
    '#type' => 'textfield',
    '#title' => t('Default value'),
    '#description' => t("If given, then this is the default value that this attribute will be set for new objects."),
    '#default_value' => $attribute['default_value'],
  );
  $form['definition']['form_display'] = array(
    '#type' => 'textfield',
    '#title' => t('Form display'),
    '#description' => t("If given, then this function will be called to determine whether to display the attribute in the object add/edit form."),
    '#default_value' => $attribute['form_display'],
  );
  $form['definition']['settings'] = module_invoke($class['module'], 'rpg_attribute_class_settings', 'settings form', $class['class'], $attribute, $type);
  $form['submit'] = array(
    '#type' => 'submit',
    '#value' => t('Submit'),
  );
  return $form;
}

function _rpg_admin_action_edit_form($type, $action) {
  $form = array();

  $form['type'] = array(
    '#type' => 'value',
    '#value' => $type['type'],
  );
  $form['action'] = array(
    '#type' => 'value',
    '#value' => $action['action'],
  );
  $form['definition'] = array(
    '#type' => 'fieldset',
    '#title' => t('Action definition'),
    '#description' => t("The following items only affect the '%action' action of the %type type, and any types derived from it.", array('%action' => $action['name'], '%type' => $type['name'])),
    '#collapsible' => true,
    '#collapsed' => false,
  );
  $form['definition']['name'] = array(
    '#type' => 'textfield',
    '#title' => t('Name'),
    '#default_value' => $action['name'],
  );
  $form['definition']['description'] = array(
    '#type' => 'textarea',
    '#title' => t('Description'),
    '#default_value' => $action['description'],
  );
  $form['definition']['script'] = array(
    '#type' => 'fieldset',
    '#title' => t('Action script'),
    '#collapsible' => true,
    '#collapsed' => false,
    '#description' => t('Insert the functions to be called here. Please follow the suggested guidelines for programming action scripts. If not, not only might you break your game, you might inadvertantly compromise the security of the rest of your site.'),
  );
  $form['definition']['script']['php'] = array(
    '#type' => 'textfield',
    '#title' => t('Action'),
    '#default_value' => $action['php'],
    '#description' => t('This function will be called when an action is performed.'),
  );
  $form['definition']['script']['expose'] = array(
    '#type' => 'textfield',
    '#title' => t('Expose'),
    '#default_value' => $action['expose'],
    '#description' => t('This function will be called to see whether to display an action as available to a character.'),
  );

  $form['submit'] = array(
    '#type' => 'submit',
    '#value' => t('Submit'),
  );
  return $form;
}

function _rpg_admin_event_edit_form($type, $event) {
  $form = array();

  $form['type'] = array(
    '#type' => 'value',
    '#value' => $type['type'],
  );
  $form['event'] = array(
    '#type' => 'value',
    '#value' => $event['event'],
  );
  $form['definition'] = array(
    '#type' => 'fieldset',
    '#title' => t('Event definition'),
    '#description' => t("The following items only affect the '%event' event of the %type type, and any types derived from it.", array('%event' => $event['name'], '%type' => $type['name'])),
    '#collapsible' => true,
    '#collapsed' => false,
  );
  $form['definition']['name'] = array(
    '#type' => 'textfield',
    '#title' => t('Name'),
    '#default_value' => $event['name'],
  );
  $form['definition']['description'] = array(
    '#type' => 'textarea',
    '#title' => t('Description'),
    '#default_value' => $event['description'],
  );
  $form['definition']['script'] = array(
    '#type' => 'fieldset',
    '#title' => t('Event script'),
    '#collapsible' => true,
    '#collapsed' => false,
    '#description' => t('Insert the functions to be called here. Please follow the suggested guidelines for programming event scripts. If not, not only might you break your game, you might inadvertantly compromise the security of the rest of your site.'),
  );
  $form['definition']['script']['php'] = array(
    '#type' => 'textfield',
    '#title' => t('Event'),
    '#default_value' => $event['php'],
    '#description' => t('This function will be called when an event is performed.'),
  );

  $form['submit'] = array(
    '#type' => 'submit',
    '#value' => t('Submit'),
  );
  return $form;
}

/**
 *  Build the form of defined & inherited attributes for a type.
 *
 *  This creates a form with two field sets, for defined and inherited attributes of a type.
 *  Each lists a short description and actions available from this screen (edit, delete, override, etc.).
 *  Underneath is a link to add a new attribute to the type.
 *
 *  @param $type
 *  The RPG Type array for the desired type.
 *  @return $form
 *  The form to display.
 */
function _rpg_admin_type_attributes_form($type) {
  $form = array();
  $form['inherited'] = array(
    '#type' => 'fieldset',
    '#title' => t('Inherited Attributes'),
    '#collapsible' => true,
    '#collapsed' => true,
    '#description' => t('These attributes are inherited in the %type type. Attributes surrounded with \'(parentheses)\' are overridden in the Defined Attributes section below.', array('%type' => $type['type'])),
  );
  $header = array(t('Attribute'), t('Parent Type'), t('Description'), t('Actions'));
  $rows = array();
  foreach ($type['attributes']['inherited'] as $attribute) {
    $cells = array();

    // if an inherited type is overruled by a defined type, put parentheses around it. otherwise, just have a link
    if (isset($type['attributes']['defined'][$attribute['attribute']])) {
      $cells[] = t('(!attribute)', array('!attribute' => l($attribute['name'], 'admin/rpg/types/type/' . $type['type'] . '/attributes/edit/' . $attribute['attribute'])));
    }
    else {
      $cells[] = l($attribute['name'], 'admin/rpg/types/type/' . $type['type'] . '/attributes/add/' . $attribute['attribute']);
    }
    $parent = rpg_types($attribute['inherited']);
    $cells[] = l($parent['name'], 'admin/rpg/types/type/' . $parent['type']);
    $cells[] = check_plain($attribute['description']);
    $actions = array();

    // if we have an override in place, call the edit link edit. otherwise, have an add link called override
    if (isset($type['attributes']['defined'][$attribute['attribute']])) {
      $actions[] =l(t('edit'), 'admin/rpg/types/type/' . $type['type'] . '/attributes/edit/' . $attribute['attribute']);
    }
    else {
      $actions[] = l(t('override'), 'admin/rpg/types/type/' . $type['type'] . '/attributes/add/' . $attribute['attribute']);
    }
    $cells[] = implode(', ', $actions);
    $rows[] = $cells;
  }
  $form['inherited']['attributes'] = array(
    '#type' => 'markup',
    '#value' => theme('table', $header, $rows),
  );

  $form['attributes'] = array(
    '#type' => 'fieldset',
    '#title' => t('Defined Attributes'),
    '#collapsible' => true,
    '#collapsed' => true,
    '#description' => t('These attributes are defined for the %type type. Attributes marked with \'*\' are overridden over their inherited values. If deleted from this section, overridden attributes will be reverted to their inherited value.', array('%type' => $type['type'])),
  );
  $header = array(t('Attribute'), t('Description'), t('Actions'));
  $rows = array();
  foreach ($type['attributes']['defined'] as $attribute) {
    $cells = array();
    if (isset($type['attributes']['inherited'][$attribute['attribute']])) {
      $cells[] = t('!attribute *', array('!attribute' => l($attribute['name'], 'admin/rpg/types/type/' . $type['type'] . '/attributes/edit/' . $attribute['attribute'])));
    }
    else {
      $cells[] = l($attribute['name'], 'admin/rpg/types/type/' . $type['type'] . '/attributes/edit/' . $attribute['attribute']);
    }
    $cells[] = check_plain($attribute['description']);
    $actions = array();
    $actions[] = l(t('edit'), 'admin/rpg/types/type/' . $type['type'] . '/attributes/edit/' . $attribute['attribute']);
    $actions[] = l(t('delete'), 'admin/rpg/types/type/' . $type['type'] . '/attributes/delete/' . $attribute['attribute']);
    $cells[] = implode(', ', $actions);
    $rows[] = $cells;
  }
  $form['attributes']['attributes'] = array(
    '#type' => 'markup',
    '#value' => theme('table', $header, $rows),
  );
  $form['add'] = array(
    '#type' => 'markup',
    '#value' => l('Add new attribute', 'admin/rpg/types/type/' . $type['type'] . '/attributes/add'),
  );
  return $form;
}

/**
 *  Build the form of defined & inherited actions for a type.
 *
 *  This creates a form with two field sets, for defined and inherited actions of a type.
 *  Each lists a short description and actions available from this screen (edit, delete, override, etc.).
 *  Underneath is a link to add a new attribute to the type.
 *
 *  @param $type
 *  The RPG Type array for the desired type.
 *  @return $form
 *  The form to display.
 */
function _rpg_admin_type_actions_form($type) {
  $form = array();
  $form['inherited'] = array(
    '#type' => 'fieldset',
    '#title' => t('Inherited actions'),
    '#collapsible' => true,
    '#collapsed' => true,
    '#description' => t('These actions are inherited in the %type type. Actions surrounded with \'(parentheses)\' are overridden in the Defined Actions section below.', array('%type' => $type['type'])),
  );
  $header = array(t('Action'), t('Parent Type'), t('Description'), t('Actions'));
  $rows = array();
  foreach ($type['actions']['inherited'] as $action) {
    $cells = array();
    if (isset($type['actions']['defined'][$action['action']])) {
      $cells[] = t('(!action)', array('!action' => l($action['name'], 'admin/rpg/types/type/' . $type['type'] . '/actions/edit/' . $action['action'])));
    }
    else {
      $cells[] = l($action['name'], 'admin/rpg/types/type/' . $type['type'] . '/actions/add/' . $action['action']);
    }
    $parent = rpg_types($action['inherited']);
    $cells[] = l($parent['name'], 'admin/rpg/types/type/' . $parent['type']);
    $cells[] = check_plain($action['description']);
    $actions = array();
    if (isset($type['actions']['defined'][$action['attribute']])) {
      $actions[] =l(t('edit'), 'admin/rpg/types/type/' . $type['type'] . '/actions/edit/' . $action['attribute']);
    }
    else {
      $actions[] = l(t('override'), 'admin/rpg/types/type/' . $type['type'] . '/actions/add/' . $action['attribute']);
    }
    $cells[] = implode(', ', $actions);
    $rows[] = $cells;
  }
  $form['inherited']['actions'] = array(
    '#type' => 'markup',
    '#value' => theme('table', $header, $rows),
  );

  $form['actions'] = array(
    '#type' => 'fieldset',
    '#title' => t('Defined actions'),
    '#collapsible' => true,
    '#collapsed' => true,
    '#description' => t('These actions are defined for the %type type. Actions marked with \'*\' are overridden over their inherited values. If deleted from this section, overridden actions will be reverted to their inherited value.', array('%type' => $type['type'])),
  );
  $header = array(t('Action'), t('Description'), t('Actions'));
  $rows = array();
  foreach ($type['actions']['defined'] as $action) {
    $cells = array();
    if (isset($type['actions']['inherited'][$action['action']])) {
      $cells[] = t('!action *', array('!action' => l($action['name'], 'admin/rpg/types/type/' . $type['type'] . '/actions/edit/' . $action['action'])));
    }
    else {
      $cells[] = l($action['name'], 'admin/rpg/types/type/' . $type['type'] . '/actions/edit/' . $action['action']);
    }
    $cells[] = check_plain($action['description']);
    $actions = array();
    $actions[] = l(t('edit'), 'admin/rpg/types/type/' . $type['type'] . '/actions/edit/' . $action['action']);
    $actions[] = l(t('delete'), 'admin/rpg/types/type/' . $type['type'] . '/actions/delete/' . $action['action']);
    $cells[] = implode(', ', $actions);
    $rows[] = $cells;
  }
  $form['actions']['actions'] = array(
    '#type' => 'markup',
    '#value' => theme('table', $header, $rows),
  );
  $form['add'] = array(
    '#type' => 'markup',
    '#value' => l('Add new action', 'admin/rpg/types/type/' . $type['type'] . '/actions/add'),
  );
  return $form;
}

/**
 *  Build the form of defined & inherited events for a type.
 *
 *  This creates a form with two field sets, for defined and inherited events of a type.
 *  Each lists a short description and events available from this screen (edit, delete, override, etc.).
 *  Underneath is a link to add a new attribute to the type.
 *
 *  @param $type
 *  The RPG Type array for the desired type.
 *  @return $form
 *  The form to display.
 */
function _rpg_admin_type_events_form($type) {
  $form = array();
  $form['inherited'] = array(
    '#type' => 'fieldset',
    '#title' => t('Inherited events'),
    '#collapsible' => true,
    '#collapsed' => true,
    '#description' => t('These events are inherited in the %type type. Events surrounded with \'(parentheses)\' are overridden in the Defined Events section below.', array('%type' => $type['type'])),
  );
  $header = array(t('Event'), t('Parent Type'), t('Description'), t('Events'));
  $rows = array();
  foreach ($type['events']['inherited'] as $event) {
    $cells = array();
    if (isset($type['events']['defined'][$event['event']])) {
      $cells[] = t('(!event)', array('!event' => l($event['name'], 'admin/rpg/types/type/' . $type['type'] . '/events/edit/' . $event['event'])));
    }
    else {
      $cells[] = l($event['name'], 'admin/rpg/types/type/' . $type['type'] . '/events/add/' . $event['event']);
    }
    $parent = rpg_types($event['inherited']);
    $cells[] = l($parent['name'], 'admin/rpg/types/type/' . $parent['type']);
    $cells[] = check_plain($event['description']);
    $events = array();
    if (isset($type['events']['defined'][$event['attribute']])) {
      $events[] =l(t('edit'), 'admin/rpg/types/type/' . $type['type'] . '/events/edit/' . $event['attribute']);
    }
    else {
      $events[] = l(t('override'), 'admin/rpg/types/type/' . $type['type'] . '/events/add/' . $event['attribute']);
    }
    $cells[] = implode(', ', $events);
    $rows[] = $cells;
  }
  $form['inherited']['events'] = array(
    '#type' => 'markup',
    '#value' => theme('table', $header, $rows),
  );

  $form['events'] = array(
    '#type' => 'fieldset',
    '#title' => t('Defined events'),
    '#collapsible' => true,
    '#collapsed' => true,
    '#description' => t('These events are defined for the %type type. Events marked with \'*\' are overridden over their inherited values. If deleted from this section, overridden events will be reverted to their inherited value.', array('%type' => $type['type'])),
  );
  $header = array(t('Event'), t('Description'), t('Events'));
  $rows = array();
  foreach ($type['events']['defined'] as $event) {
    $cells = array();
    if (isset($type['events']['inherited'][$event['event']])) {
      $cells[] = t('!event *', array('!event' => l($event['name'], 'admin/rpg/types/type/' . $type['type'] . '/events/edit/' . $event['event'])));
    }
    else {
      $cells[] = l($event['name'], 'admin/rpg/types/type/' . $type['type'] . '/events/edit/' . $event['event']);
    }
    $cells[] = check_plain($event['description']);
    $events = array();
    $events[] = l(t('edit'), 'admin/rpg/types/type/' . $type['type'] . '/events/edit/' . $event['event']);
    $events[] = l(t('delete'), 'admin/rpg/types/type/' . $type['type'] . '/events/delete/' . $event['event']);
    $cells[] = implode(', ', $events);
    $rows[] = $cells;
  }
  $form['events']['events'] = array(
    '#type' => 'markup',
    '#value' => theme('table', $header, $rows),
  );
  $form['add'] = array(
    '#type' => 'markup',
    '#value' => l('Add new event', 'admin/rpg/types/type/' . $type['type'] . '/events/add'),
  );
  return $form;
}

/**
 *  Build the form to add (or override) an attribute for an RPG type.
 *
 *  This form will include all the elements for an attribute of the given class. If overriding a type, then the elements
 *  will be prepopulated with the inherited values.
 *
 *  @param $type
 *  The RPG type array to add the array to, as called by rpg_types($type);
 *  @param $attribute
 *  An optional string. If given, then we will override the attribute defined by the type's parent(s).
 */
function _rpg_admin_type_attribute_add($type, $attribute = NULL) {
  // allow a type to override an inherited attribute, or define an attribute from outside its tree
  if (isset($attribute)) {
    $attribute = rpg_attributes($attribute);
    if (!isset($attribute['attribute'])) {
      drupal_not_found();
    }
    // add the attribute to be added or overridden
    if (!isset($type['attributes']['defined'][$attribute['attribute']])) {
      if (isset($type['attributes']['inherited'][$attribute['attribute']])) {
        $override = true;
        drupal_set_message(t('Overriding the existing %attribute attribute for the %type type. You may edit it below.', array('%attribute' => $attribute['name'], '%type' => $type['name'])));
      }
      else {
        $override = false;
        drupal_set_message(t('Added new %attribute attribute to the %type type. You may edit it below.', array('%attribute' => $attribute['name'], '%type' => $type['name'])));
      }
      // include rpg_attribute_add
      rpg_include_loadcache();

      $attribute = rpg_attribute_add($type['type'], $attribute);
    }
    else {
      drupal_set_message(t('The %attribute attribute is already defined for the %type type. However, you may edit it below.', array('%attribute' => $attribute['name'], '%type' => $type['name'])), 'error');
    }
    // redirect to the attribute's new edit screen
    drupal_goto('admin/rpg/types/type/' . $type['type'] . '/attributes/edit/' . $attribute['attribute']);
  }
  else {
    $attribute = array();
    $attribute['is_new'] = true;
  }

  return _rpg_admin_attribute_add_form($type, $attribute);
}

/**
 *  Build the form to add (or override) an action for an RPG type.
 *
 *  This form will include all the elements for an action of the given class. If overriding a type, then the elements
 *  will be prepopulated with the inherited values.
 *
 *  @param $type
 *  The RPG type array to add the action to, as called by rpg_types($type);
 *  @param $action
 *  An optional string. If given, then we will override the action defined by the type's parent(s).
 */
function _rpg_admin_type_action_add($type, $action = NULL) {
  // allow a type to override an inherited action, or define an action from outside its tree
  if (isset($action)) {
    $action = rpg_actions($action);
    if (!isset($action['action'])) {
      drupal_not_found();
    }
    // add the action to be added or overridden
    if (!isset($type['actions']['defined'][$attribute['action']])) {
      if (isset($type['actions']['inherited'][$attribute['action']])) {
        $override = true;
        drupal_set_message(t('Overriding the existing %action action for the %type type. You may edit it below.', array('%action' => $action['name'], '%type' => $type['name'])));
      }
      else {
        $override = false;
        drupal_set_message(t('Added new %action action to the %type type. You may edit it below.', array('%action' => $action['name'], '%type' => $type['name'])));
      }
      // include rpg_action_add
      rpg_include_loadcache();

      $action = rpg_action_add($type['type'], $action['action'], $action['name']);
    }
    else {
      drupal_set_message(t('The %action action is already defined for the %type type. However, you may edit it below.', array('%action' => $action['name'], '%type' => $type['name'])), 'error');
    }
    // redirect to the action's new edit screen
    drupal_goto('admin/rpg/types/type/' . $type['type'] . '/actions/edit/' . $action['action']);
  }
  else {
    $action = array();
    $action['is_new'] = true;
  }

  return _rpg_admin_action_add_form($type, $action);
}


/**
 *  Build the form to add (or override) an event for an RPG type.
 *
 *  This form will include all the elements for an event of the given class. If overriding a type, then the elements
 *  will be prepopulated with the inherited values.
 *
 *  @param $type
 *  The RPG type array to add the event to, as called by rpg_types($type);
 *  @param $event
 *  An optional string. If given, then we will override the event defined by the type's parent(s).
 */
function _rpg_admin_type_event_add($type, $event = NULL) {
  // allow a type to override an inherited event, or define an event from outside its tree
  if (isset($event)) {
    $event = rpg_events($event);
    if (!isset($event['event'])) {
      drupal_not_found();
    }
    // add the event to be added or overridden
    if (!isset($type['events']['defined'][$attribute['event']])) {
      if (isset($type['events']['inherited'][$attribute['event']])) {
        $override = true;
        drupal_set_message(t('Overriding the existing %event event for the %type type. You may edit it below.', array('%event' => $event['name'], '%type' => $type['name'])));
      }
      else {
        $override = false;
        drupal_set_message(t('Added new %event event to the %type type. You may edit it below.', array('%event' => $event['name'], '%type' => $type['name'])));
      }
      // include rpg_event_add
      rpg_include_loadcache();

      $event = rpg_event_add($type['type'], $event['event'], $event['name']);
    }
    else {
      drupal_set_message(t('The %event event is already defined for the %type type. However, you may edit it below.', array('%event' => $event['name'], '%type' => $type['name'])), 'error');
    }
    // redirect to the event's new edit screen
    drupal_goto('admin/rpg/types/type/' . $type['type'] . '/events/edit/' . $event['event']);
  }
  else {
    $event = array();
    $event['is_new'] = true;
  }

  return _rpg_admin_event_add_form($type, $event);
}
